#include "lion/ECT.h"
#include "lion/LPP.h"
#include "lion/rtos/RTOS_Mutex.h"
#include "lion/log/log.h"
#include "ws.h"
#include <ctype.h>
#include <string.h>
#include <stdio.h>
#include <avahi-client/client.h>
#include <avahi-client/publish.h>
#include <avahi-common/thread-watch.h>
#include <avahi-common/error.h>

LOG_MODULE_REGISTER(ect, LOG_LEVEL_DBG);

/**********************************************************************************************************************
 * @cond                                               FILE SCOPE                                              @endcond
 *********************************************************************************************************************/

FileScop AvahiClient *ect_ptr_AvahiClient;
FileScop AvahiThreadedPoll *ect_ptr_AvahiThreadedPoll;
FileScop AvahiEntryGroup *ect_ptr_AvahiEntryGroup = NULL;
FileScop LPP_Transport_t ect_str_Transport;

/**
 * @brief Callback for sending messages back to client
 * @ref LPP_TransportTX_t
 */
int ect_TransportTX(void *ptr_Destination, const uint8_t *ptr_Data, size_t siz_Length, RTOS_Timeout_t tim_Timeout, void *ptr_UserData) {
    ws_cli_conn_t *ptr_Client = (ws_cli_conn_t *)ptr_Destination;
    ws_sendframe(ptr_Client, ptr_Data, siz_Length, WS_FR_OP_TXT);
}

/* The following are intended to be implemented, but the LPP doesn't support it yet.
ssize_t ect_TransportRX(void *ptr_Source, uint8_t *ptr_Data, size_t siz_Length, RTOS_Timeout_t tim_Timeout, void *ptr_UserData) {}
size_t ect_TransportGetMTU(void *ptr_Source, void *ptr_UserData) {}
*/

/**
 * @brief Callback for @c ws_events : On open event, called when a new client connects.
 * @param ptr_Client WebSocket Client Reference
 */
FileScop void ect_OnOpen(ws_cli_conn_t *ptr_Client)
{
    LOG_WRN("OnOpen(%s)", ws_getaddress(ptr_Client));
}

/**
 * @brief Callback for @c ws_events : On close event, called when a client disconnects.
 * @param ptr_Client WebSocket Client Reference
 */
FileScop void ect_OnClose(ws_cli_conn_t *ptr_Client)
{
    LOG_WRN("OnClose(%s)", ws_getaddress(ptr_Client));
}

/**
 * @brief Callback for @c ws_events : On message event, called when a client sends a text or binary message.
 * @param ptr_Client WebSocket Client Reference
 * @param msg Message data
 * @param size Message Length 
 * @param type WebSocket message type: WS_FR_OP_TXT, WS_FR_OP_BIN, etc.
 */
FileScop void ect_OnMessage(ws_cli_conn_t *ptr_Client,
                            const unsigned char *msg, uint64_t size, int type)
{
    LOG_WRN("OnMessage(%s, %llu, %d)", ws_getaddress(ptr_Client), size, type);
    LPP_RXFrame(&ect_str_Transport, ptr_Client, msg, size);
}

FileScop struct ws_events ect_str_Events = {
    .onopen = ect_OnOpen,
    .onclose = ect_OnClose,
    .onmessage = ect_OnMessage};


/**
 * @brief Callback for @c avahi_entry_group_new : Called when the group state changes
 * @param ptr_Group Avahi entry group object
 * @param enm_State Current group state
 * @param ptr_UserData 
 */
FileScop void ect_AvahiGroupCallback(AvahiEntryGroup *ptr_Group, AvahiEntryGroupState enm_State, void *ptr_UserData)
{
    switch (enm_State)
    {
    case AVAHI_ENTRY_GROUP_ESTABLISHED:
        LOG_INF("Service successfully established.");
        break;
    case AVAHI_ENTRY_GROUP_COLLISION:

    case AVAHI_ENTRY_GROUP_FAILURE:

    case AVAHI_ENTRY_GROUP_UNCOMMITED:

    case AVAHI_ENTRY_GROUP_REGISTERING:
    }
}

/**
 * @brief Called to initialize the Avahi group
 * @param ptr_Client Avahi Connection Context
 */
FileScop void ect_AvahiCreateService(AvahiClient *ptr_Client)
{
    if (!ect_ptr_AvahiEntryGroup)
    {
        ect_ptr_AvahiEntryGroup = avahi_entry_group_new(ptr_Client, ect_AvahiGroupCallback, NULL);
        if (!ect_AvahiGroupCallback)
        {
            LOG_ERR("avahi_entry_group_new() failed: %s", avahi_strerror(avahi_client_errno(ptr_Client)));
            return; // TODO: Maybe completely abort?
        }
    }

    /// @todo This should probably be moved somewhere globally accessible
    FuncScop char sz_ServiceName[128];
    FuncScop char sz_RecordSN[128];
    StackScop char sz_SerialNumber[48];
    FILE *ptr_SerialNumber = fopen("/proc/device-tree/serial-number","r");
    if (ptr_SerialNumber != NULL) {
        StackScop size_t siz_Read = fread(sz_SerialNumber, 1, sizeof(sz_SerialNumber)-1, ptr_SerialNumber);
        sz_SerialNumber[siz_Read] = '\0'; /* Just to be safe. */
        for (int i = 0; i < strlen(sz_SerialNumber); i++)
        {
            sz_SerialNumber[i] = toupper(sz_SerialNumber[i]);
        }
        fclose(ptr_SerialNumber);
    } else {
        strcpy(sz_SerialNumber, "UNKNOWN");
    }
    snprintf(sz_ServiceName, sizeof(sz_ServiceName), "Lion EMS (%s)", sz_SerialNumber);
    snprintf(sz_RecordSN, sizeof(sz_RecordSN), "sn=%s", sz_SerialNumber);

    if (avahi_entry_group_is_empty(ect_ptr_AvahiEntryGroup))
    {
        StackScop int err_RC;
        LOG_INF("Adding Avahi Services");

        err_RC = avahi_entry_group_add_service(ect_ptr_AvahiEntryGroup, AVAHI_IF_UNSPEC, AVAHI_PROTO_UNSPEC, 0, sz_ServiceName, "_lion-energy._tcp", NULL, NULL, 8080, sz_RecordSN, NULL);
        if (err_RC < 0)
        {
            LOG_ERR("avahi_entry_group_add_service() failed: %s", avahi_strerror(avahi_client_errno(ptr_Client)));
            return;
        }

        err_RC = avahi_entry_group_add_service_subtype(ect_ptr_AvahiEntryGroup, AVAHI_IF_UNSPEC, AVAHI_PROTO_UNSPEC, 0, sz_ServiceName, "_lion-energy._tcp", NULL, "_http._sub._lion-energy._tcp");
        if (err_RC < 0)
        {
            LOG_ERR("avahi_entry_group_add_service_subtype() failed: %s", avahi_strerror(avahi_client_errno(ptr_Client)));
        }

        err_RC = avahi_entry_group_commit(ect_ptr_AvahiEntryGroup);
        if (err_RC < 0)
        {
            LOG_ERR("avahi_entry_group_commit() failed: %s", avahi_strerror(avahi_client_errno(ptr_Client)));
            return;
        }
    }
}

/**
 * @brief Callback for @c avahi_client_new : Called whenever the Avahi client state changes
 * @param ptr_Client Avahi Connection Context
 * @param enm_State 
 * @param ptr_UserData
 */
FileScop void ect_AvahiClientCallback(AvahiClient *ptr_Client, AvahiClientState enm_State, void *ptr_UserData)
{
    switch (enm_State)
    {
    case AVAHI_CLIENT_S_RUNNING:
        LOG_INF("Avahi Running");
        /* The server has startup successfully and registered its host
         * name on the network, so it's time to create our services */
        ect_AvahiCreateService(ptr_Client);
        break;
    case AVAHI_CLIENT_FAILURE:
        LOG_ERR("Avahi failure: %s\n", avahi_strerror(avahi_client_errno(ptr_Client)));
        break;
    case AVAHI_CLIENT_S_COLLISION:
        LOG_WRN("Avahi Service Collision");
        break;
    case AVAHI_CLIENT_S_REGISTERING:
        LOG_INF("Avahi Service Registering");
        if (ect_ptr_AvahiEntryGroup)
        {
            avahi_entry_group_reset(ect_ptr_AvahiEntryGroup);
        }
        break;
    case AVAHI_CLIENT_CONNECTING:
        LOG_INF("Avahi Connecting");
        break;
    default:
        break;
    }
}

/**********************************************************************************************************************
 * @cond                                              GLOBAL SCOPE                                             @endcond
 *********************************************************************************************************************/

GlobScop void ECT_Init()
{
    int err_RC;

    LPP_InitTransport(&ect_str_Transport, ect_TransportTX, NULL);

    ect_ptr_AvahiThreadedPoll = avahi_threaded_poll_new();
    ect_ptr_AvahiClient = avahi_client_new(avahi_threaded_poll_get(ect_ptr_AvahiThreadedPoll), AVAHI_CLIENT_NO_FAIL, ect_AvahiClientCallback, NULL, &err_RC);
    if (!ect_ptr_AvahiClient)
    {
        LOG_ERR("Failed to create Avahi client: %s", avahi_strerror(err_RC));
    }
}

GlobScop int ECT_Start()
{
    LPP_RegisterTransport(&ect_str_Transport);
    ws_socket(&ect_str_Events, 8080, 1, 1000);
    avahi_threaded_poll_start(ect_ptr_AvahiThreadedPoll);
}

GlobScop int ECT_Stop(RTOS_Timeout_t tim_Timeout)
{
    avahi_threaded_poll_stop(ect_ptr_AvahiThreadedPoll);
}
