#include <lion/PMS.h>
#include <lion/ial/ccu/IAL_Modbus.h>
#include <lion/log/log.h>
#include <lion/rtos/RTOS_Sleep.h>
#include <errno.h>
#include <modbus/modbus.h>
#include <stdlib.h>
#include <zcbor_decode.h>
#include <zcbor_encode.h>

#define pms_LOG_LVL LOG_LEVEL_INF
LOG_MODULE_REGISTER(pms, pms_LOG_LVL);

/*********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

//---------------------------------------//
// Section: Defines/Enums/Typedefs       //
//---------------------------------------//

#define pms_SLAVE_UNIT_TYPE        COD_sz_UNIT_TYPE_PCS
#define pms_sz_RTU_IFACE           "/dev/ttyUSB0"
#define pms_u32_RTU_BAUDRATE       19200
#define pms_u8_RTU_PARITY          'N'
#define pms_u8_RTU_DATA_BITS       8
#define pms_u8_RTU_STOP_BITS       1
#define pms_u8_RTU_SLAVE_ADDR      1

#define pms_MAX_READ_REGS          MODBUS_MAX_READ_REGISTERS

#define pms_IDLE_SLEEP_TIME_MS     100
#define pms_DISCONN_SLEEP_TIME_MS  5000
#define pms_NCONN_SLEEP_TIME_MS    (15U * 60U * 1000U) //15 minutes

#define MODBUS_RTU_HEADER_LENGTH 1
#define MODBUS_SLAVE(slave, req) uint8_t slave = req[MODBUS_RTU_HEADER_LENGTH - 1];
#define MODBUS_FUNC_CODE(fc, req) uint8_t fc = req[MODBUS_RTU_HEADER_LENGTH];
#define MODBUS_READ_NUM(nb, req) size_t nb = (req[MODBUS_RTU_HEADER_LENGTH + 3] << 8) | req[MODBUS_RTU_HEADER_LENGTH + 4];
#define MODBUS_READ_ADDR(addr, req) \
    uint16_t addr = (req[MODBUS_RTU_HEADER_LENGTH + 1] << 8) | req[MODBUS_RTU_HEADER_LENGTH + 2]; \
    do { req[MODBUS_RTU_HEADER_LENGTH + 1] = req[MODBUS_RTU_HEADER_LENGTH + 2] = 0; } while (0)

typedef enum pms_state
{
    STATE_IDLE,
    STATE_PROCESS_REQ,
    STATE_DISCONNECTED,
    STATE_ERROR
} PmsState_t;

//---------------------------------------//
// Section: Private Variables/Constants  //
//---------------------------------------//

FileScop pthread_t str_PmsThread;
FileScop RTOS_Mutex_t pms_str_Mutex; 
FileScop modbus_t *ptr_str_ModbusCtx = NULL; //Context structure used by libmodbus
FileScop modbus_mapping_t *ptr_str_ModbusMapping = NULL; //libmodbus mapping structure. Points to the temporary registers.

FileScop bool bol_Initialized = false; //Flag to indicate if the application has been initialized.
FileScop bool bol_Running = false; //Flag to indicate if the modbus server is running.

//---------------------------------------//
// Section: Function Prototypes          //
//---------------------------------------//

/**
 * @brief The entry point for the PMS thread.
 */
FileScop void *pms_ThreadEntry(void *arg);

/**
 * @brief Starts the modbus server.
 * 
 * @returns EXIT_SUCCESS or 0 on success, libmodbus error code on failure.
 */
FileScop int pms_StartModbusServer(void);

/**
 * @brief Stops the modbus server.
 * 
 * @returns EXIT_SUCCESS or 0 on success, libmodbus error code on failure.
 */
FileScop int pms_StopModbusServer(void);

/**
 * @brief Handles a modbus request from a client.
 * 
 * @param ptr_str_ModbusCtx The modbus context.
 * @param arr_u8_Query The modbus request.
 * @param int_ReqLen The length of the request.
 * 
 * @returns EXIT_SUCCESS or 0 on success, libmodbus error code on failure.
 */
FileScop int pms_HandleRequest(modbus_t *ptr_str_ModbusCtx, uint8_t *arr_u8_Query, int int_ReqLen);

/**
 * @brief Wrapper function for reading holding registers from the ICC.
 * 
 * @details Calls @ref pms_ReadPcsRegs with the destination registers set to the modbus mapping holding registers.
 * 
 * @param ptr_str_ModbusCtx The modbus context.
 * @param arr_u8_Query The modbus request.
 * @param int_ReqLen The length of the request.
 * 
 * @returns EXIT_SUCCESS or 0 on success, libmodbus error code on failure.
 */
FileScop inline int pms_ReadHoldingRegs(modbus_t *ptr_str_ModbusCtx, uint8_t *arr_u8_Query, int int_ReqLen);

/**
 * @brief Wrapper function for reading input registers from the ICC.
 * 
 * @details Calls @ref pms_ReadPcsRegs with the destination registers set to the modbus mapping input registers.
 * 
 * @param ptr_str_ModbusCtx The modbus context.
 * @param arr_u8_Query The modbus request.
 * @param int_ReqLen The length of the request.
 * 
 * @returns EXIT_SUCCESS or 0 on success, libmodbus error code on failure.
 */
FileScop inline int pms_ReadInputRegs(modbus_t *ptr_str_ModbusCtx, uint8_t *arr_u8_Query, int int_ReqLen);

/**
 * @brief Reads holding registers from the ICC and sends the response to the client.
 * 
 * @note This function will modify the query buffer by setting the requested address to zero.
 *       Doing so allows us to both read through the ICC and send the response using libmodbus APIs.
 * 
 * @param ptr_str_ModbusCtx The modbus context.
 * @param arr_u8_Query The modbus request.
 * @param int_ReqLen The length of the request.
 * @param ptr_u16_DestRegs The destination registers to store the read values.
 * 
 * @returns EXIT_SUCCESS or 0 on success, libmodbus error code on failure.
 */
FileScop int pms_ReadPcsRegs(modbus_t *ptr_str_ModbusCtx, uint8_t *arr_u8_Query, int int_ReqLen, uint16_t *ptr_u16_DestRegs);


/*********************************************************************************************************************
* @cond                                              GLOBAL SCOPE                                             @endcond
*********************************************************************************************************************/

GlobScop int PMS_Init(void)
{
    if (bol_Initialized)
    {
        return -EALREADY;
    }

    RTOS_MutexInit(&pms_str_Mutex);

    bol_Initialized = true;
    return EXIT_SUCCESS;
}

GlobScop int PMS_Start(void)
{
    StackScop int int_Rc = EXIT_SUCCESS;
    RTOS_MutexLock(&pms_str_Mutex, RTOS_FOREVER);
    if (!bol_Initialized)
    {
        int_Rc = -EPERM;
    }
    else if (bol_Running)
    {
        int_Rc = -EALREADY;
    }
    else
    {
        pthread_create(&str_PmsThread, NULL, &pms_ThreadEntry, NULL);
        bol_Running = true;
    }
    RTOS_MutexUnlock(&pms_str_Mutex);
    return int_Rc;
}

GlobScop int PMS_Stop(RTOS_Timeout_t tim_Timeout)
{
    StackScop int int_Rc = EXIT_SUCCESS;
    RTOS_MutexLock(&pms_str_Mutex, RTOS_FOREVER);
    if (!bol_Running)
    {
        int_Rc = -EALREADY;
    }
    else
    {
        bol_Running = false;
        pthread_join(str_PmsThread, NULL);
        pms_StopModbusServer();
    }
    RTOS_MutexUnlock(&pms_str_Mutex);
    return int_Rc;
}

/*********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

FileScop void *pms_ThreadEntry(void *arg)
{
    FuncScop PmsState_t enm_State = STATE_DISCONNECTED;
    FuncScop uint8_t arr_u8_Query[MODBUS_RTU_MAX_ADU_LENGTH];
    FuncScop int int_ReqLen = 0;
    FuncScop int int_Err = 0;
    FuncScop bool bol_WasConnected = false;

    while (bol_Running)
    {
        switch (enm_State)
        {
            case STATE_IDLE:
                int_ReqLen = modbus_receive(ptr_str_ModbusCtx, arr_u8_Query);
                if (int_ReqLen > 0)
                {
                    enm_State = STATE_PROCESS_REQ;
                }
                else if (int_ReqLen == -1)
                {
                    int_Err = -errno;
                    LOG_ERR("Failed to receive modbus request, err %d (%s)", int_Err, modbus_strerror(-int_Err));
                    if (int_Err == -ECONNRESET)
                    {
                        pms_StopModbusServer();
                        enm_State = STATE_DISCONNECTED;
                    }
                    else
                    {
                        enm_State = STATE_ERROR;
                    }
                }
                else
                {
                    RTOS_Sleep(RTOS_MSEC(pms_IDLE_SLEEP_TIME_MS));
                    enm_State = STATE_IDLE;
                }
                break;
            case STATE_PROCESS_REQ:
                int_Err = pms_HandleRequest(ptr_str_ModbusCtx, arr_u8_Query, int_ReqLen);
                if (int_Err != EXIT_SUCCESS)
                {
                    LOG_ERR("Failed to handle modbus request, err %d (%s)", int_Err, modbus_strerror(-int_Err));
                    enm_State = STATE_ERROR;
                }
                else
                {
                    enm_State = STATE_IDLE;
                }
                break;
            case STATE_DISCONNECTED:
                RTOS_Sleep(RTOS_MSEC(pms_DISCONN_SLEEP_TIME_MS));
                int_Err = pms_StartModbusServer();
                if (int_Err == EXIT_SUCCESS)
                {
                    bol_WasConnected = true;
                    enm_State = STATE_IDLE;
                }
                else if (!bol_WasConnected)
                {
                    //In the case of a system that does not have any client device connected, no RTU interface will be 
                    //available and the initial connection attempt will fail. Wait for several minutes before checking 
                    //for a connection again to decrease CPU usage.
                    LOG_WRN("RTU interface not detected, going to sleep...");
                    RTOS_Sleep(RTOS_MSEC(pms_NCONN_SLEEP_TIME_MS));
                }
                break;
            case STATE_ERROR:
              /* Fall through */
            default:
                int_Err = 0;
                enm_State = STATE_IDLE;
                break;
        }
    }

    return NULL;
}

FileScop int pms_StartModbusServer(void)
{
    if (ptr_str_ModbusCtx != NULL)
    {
        return -EALREADY;
    }

    ptr_str_ModbusCtx = modbus_new_rtu(pms_sz_RTU_IFACE, pms_u32_RTU_BAUDRATE, pms_u8_RTU_PARITY, 
                                        pms_u8_RTU_DATA_BITS, pms_u8_RTU_STOP_BITS);
    if (ptr_str_ModbusCtx == NULL)
    {
        LOG_ERR("Failed to create modbus context, err -%d (%s)", errno, modbus_strerror(errno));
        return -errno;
    }

    modbus_set_debug(ptr_str_ModbusCtx, (pms_LOG_LVL == LOG_LEVEL_DBG) ? true : false);

    if (modbus_set_slave(ptr_str_ModbusCtx, pms_u8_RTU_SLAVE_ADDR) == -1)
    {
        LOG_ERR("Failed to set modbus slave address, err -%d (%s)", errno, modbus_strerror(errno));
        modbus_free(ptr_str_ModbusCtx);
        ptr_str_ModbusCtx = NULL;
        return -errno;
    }
    if (modbus_connect(ptr_str_ModbusCtx) == -1)
    {
        LOG_ERR("Failed to connect to modbus device, err -%d (%s)", errno, modbus_strerror(errno));
        modbus_free(ptr_str_ModbusCtx);
        ptr_str_ModbusCtx = NULL;
        return -errno;
    }
    ptr_str_ModbusMapping = modbus_mapping_new(0, 0, pms_MAX_READ_REGS, pms_MAX_READ_REGS);
    if (ptr_str_ModbusMapping == NULL)
    {
        LOG_ERR("Failed to create modbus mapping, err -%d (%s)", errno, modbus_strerror(errno));
        modbus_free(ptr_str_ModbusCtx);
        ptr_str_ModbusCtx = NULL;
        return -errno;
    }
    return EXIT_SUCCESS;
}

FileScop int pms_StopModbusServer(void)
{
    if (ptr_str_ModbusCtx == NULL)
    {
        return -EALREADY;
    }

    modbus_mapping_free(ptr_str_ModbusMapping);
    ptr_str_ModbusMapping = NULL;
    modbus_close(ptr_str_ModbusCtx);
    modbus_free(ptr_str_ModbusCtx);
    ptr_str_ModbusCtx = NULL;
    return EXIT_SUCCESS;
}

FileScop int pms_HandleRequest(modbus_t *ptr_str_ModbusCtx, uint8_t *arr_u8_Query, int int_ReqLen)
{
    StackScop int int_Rc = EXIT_SUCCESS;
    MODBUS_FUNC_CODE(u8_FunctionCode, arr_u8_Query);
    switch (u8_FunctionCode)
    {
        case MODBUS_FC_READ_HOLDING_REGISTERS:
            int_Rc = pms_ReadHoldingRegs(ptr_str_ModbusCtx, arr_u8_Query, int_ReqLen);
            break;
        case MODBUS_FC_READ_INPUT_REGISTERS:
            int_Rc = pms_ReadInputRegs(ptr_str_ModbusCtx, arr_u8_Query, int_ReqLen);
            break;
        case MODBUS_FC_READ_COILS:
        case MODBUS_FC_READ_DISCRETE_INPUTS:
        case MODBUS_FC_WRITE_SINGLE_COIL:
        case MODBUS_FC_WRITE_SINGLE_REGISTER:
        case MODBUS_FC_WRITE_MULTIPLE_COILS:
        case MODBUS_FC_WRITE_MULTIPLE_REGISTERS:
        case MODBUS_FC_REPORT_SLAVE_ID:
        case MODBUS_FC_READ_EXCEPTION_STATUS:
        case MODBUS_FC_MASK_WRITE_REGISTER:
        case MODBUS_FC_WRITE_AND_READ_REGISTERS:
            /* Fall through */
        default:
            LOG_DBG("Unsupported function code %d", u8_FunctionCode);
            int_Rc = modbus_reply_exception(ptr_str_ModbusCtx, arr_u8_Query, MODBUS_EXCEPTION_ILLEGAL_FUNCTION);
            if (int_Rc != EXIT_SUCCESS)
            {
                LOG_INF("Modbus reply failed, err -%d (%s)", errno, modbus_strerror(errno));
                int_Rc = -errno;
            }
            break;
    } /* switch(fc) */
    return int_Rc;
}

FileScop inline int pms_ReadHoldingRegs(modbus_t *ptr_str_ModbusCtx, uint8_t *arr_u8_Query, int int_ReqLen)
{
    return pms_ReadPcsRegs(ptr_str_ModbusCtx, arr_u8_Query, int_ReqLen, ptr_str_ModbusMapping->tab_registers);
}

FileScop inline int pms_ReadInputRegs(modbus_t *ptr_str_ModbusCtx, uint8_t *arr_u8_Query, int int_ReqLen)
{
    return pms_ReadPcsRegs(ptr_str_ModbusCtx, arr_u8_Query, int_ReqLen, ptr_str_ModbusMapping->tab_input_registers);
}

FileScop int pms_ReadPcsRegs(modbus_t *ptr_str_ModbusCtx, uint8_t *arr_u8_Query, int int_ReqLen, uint16_t *ptr_u16_DestRegs)
{
    MODBUS_FUNC_CODE(u8_FunctionCode, arr_u8_Query);
    MODBUS_READ_NUM(siz_ReadQty, arr_u8_Query);
    MODBUS_READ_ADDR(u16_Address, arr_u8_Query);

    LOG_INF("Received read request from client: start_address %d, quantity %d", u16_Address, siz_ReadQty);

    StackScop int int_Rc = IAL_Mbs_ReadMultipleRegisters(pms_SLAVE_UNIT_TYPE, COD_u8_NULL_UNIT_ID, u16_Address,
                                                            ptr_u16_DestRegs, siz_ReadQty);
    if (int_Rc != EXIT_SUCCESS)
    {
        LOG_INF("Failed to read multiple registers, err %d (%s)", int_Rc, strerror(-int_Rc));
        int_Rc = modbus_reply_exception(ptr_str_ModbusCtx, arr_u8_Query, MODBUS_EXCEPTION_SLAVE_OR_SERVER_BUSY);
    }
    else
    {
        LOG_DBG("Read %d registers at %d", siz_ReadQty, u16_Address);
        for (int i = 0; i < siz_ReadQty; i++)
        {
            LOG_DBG("\t%d: 0x%04X (%d)", u16_Address + i, ptr_u16_DestRegs[i], ptr_u16_DestRegs[i]);
        }

        int_Rc = modbus_reply(ptr_str_ModbusCtx, arr_u8_Query, int_ReqLen, ptr_str_ModbusMapping);
    }
    if (int_Rc < 0)
    {
        LOG_INF("Modbus reply failed, err -%d (%s)", errno, modbus_strerror(errno));
        int_Rc = -errno;
    }
    return (int_Rc < 0) ? int_Rc : EXIT_SUCCESS;
}

//EOF
