#include "ICC_Internal.h"
#include "lion/icc/ICC.h"
#include "lion/log/log.h"
#include "lion/utils/byteorder.h"
#include "lion/utils/crc.h"
#include <errno.h>
#include <string.h>
#ifndef __ZEPHYR__
    #include <stdlib.h>
#endif

LOG_MODULE_DECLARE(icc, CONFIG_ICC_LOG_LEVEL);

/**********************************************************************************************************************
 * @cond                                               FILE SCOPE                                              @endcond
 *********************************************************************************************************************/

typedef struct
{
    RTOS_Sem_t strSem;
    void      *ptrResponse;
    uint32_t  *u32ResponseLength;
    int32_t    i32ReturnValue;
} icc_Function_SyncCallback_t;

/** @brief Implementation of @ref ICC_Function_Callback_t to implement @ref ICC_Function_Call using
 * @ref ICC_Function_CallAsync . */
FileScop void icc_Function_SyncCallback(uint32_t u32FunctionId, int intResult, const void *ptrResponse,
                                        uint32_t u32ResponseLength, void *ptrUserData);

/**********************************************************************************************************************
 * @cond                                              GLOBAL SCOPE @endcond
 *********************************************************************************************************************/

GlobScop void ICC_Function_InitHandler(ICC_Function_Handler_t *handler, uint32_t u32FunctionId,
                                       uint32_t u32FunctionIdMask, ICC_Function_HandlerCallback_t callback,
                                       void *userData)
{
    sys_dnode_init(&handler->node);
    handler->u32FunctionId     = u32FunctionId;
    handler->u32FunctionIdMask = u32FunctionIdMask;
    handler->fnCallback        = callback;
    handler->ptrUserData       = userData;
}

GlobScop int ICC_Function_RegisterHandler(ICC_Context_t *context, ICC_Function_Handler_t *handler)
{
    int errcode = 0;
    if (sys_dnode_is_linked(&handler->node))
    {
        errcode = -EALREADY;
    }
    else
    {
        RTOS_MutexLock(&context->strHandlerLock, RTOS_FOREVER);
        ICC_Function_Handler_t *node;
        SYS_DLIST_FOR_EACH_CONTAINER(&context->functions, node, node)
        {
            uint32_t sharedMask = node->u32FunctionIdMask & handler->u32FunctionIdMask;
            if ((node->u32FunctionId & sharedMask) == (handler->u32FunctionId & sharedMask))
            {
                LOG_ERR("Handler Conflict 0x%08u[0x%08u] 0x%08u[0x%08u]", node->u32FunctionId, node->u32FunctionIdMask,
                        handler->u32FunctionId, handler->u32FunctionIdMask);
                errcode = -EALREADY;
                break;
            }
        }
        if (errcode == 0)
        {
            sys_dlist_append(&context->functions, &handler->node);
        }
        RTOS_MutexUnlock(&context->strHandlerLock);
    }
    return errcode;
}

GlobScop int ICC_Function_UnregisterHandler(ICC_Context_t *context, ICC_Function_Handler_t *handler)
{
    RTOS_MutexLock(&context->strHandlerLock, RTOS_FOREVER);
    sys_dlist_remove(&handler->node);
    RTOS_MutexUnlock(&context->strHandlerLock);
    return 0;
}

GlobScop int ICC_Function_Call(ICC_Context_t *ptrContext, uint32_t u32FunctionId, const void *ptrData,
                               uint32_t u32DataLength, void *ptrResponse, uint32_t *u32ResponseLength,
                               RTOS_Timeout_t strTimeout)
{
    /* u32ResponseLength MUST not be NULL if ptrResponse is not NULL */
    assert((ptrResponse == NULL) || (u32ResponseLength != NULL));

    StackScop ICC_FunctionCall_t          strCallContext;
    StackScop icc_Function_SyncCallback_t strSyncCallback;
    StackScop int                         int_RC;
    RTOS_SemInit(&strSyncCallback.strSem, 0, 1);
    strSyncCallback.ptrResponse       = ptrResponse;
    strSyncCallback.u32ResponseLength = u32ResponseLength;
    ICC_Function_InitCall(&strCallContext);
    int_RC = ICC_Function_CallAsync(ptrContext, &strCallContext, u32FunctionId, ptrData, u32DataLength,
                                    &icc_Function_SyncCallback, &strSyncCallback);
    if (int_RC < 0)
    {
        if (u32ResponseLength != NULL)
        {
            *u32ResponseLength = 0;
        }
        return int_RC;
    }

    int_RC = RTOS_SemTake(&strSyncCallback.strSem, strTimeout);
    ICC_Function_DestroyCall(&strCallContext);
    if (int_RC < 0)
    {
        if (u32ResponseLength != NULL)
        {
            *u32ResponseLength = 0;
        }
        return int_RC;
    }
    else
    {
        return strSyncCallback.i32ReturnValue;
    }
}

GlobScop int ICC_Function_CallAsync(ICC_Context_t *ptrContext, ICC_FunctionCall_t *ptrFunctionCall,
                                    uint32_t u32FunctionId, const void *ptrData, uint32_t u32DataLength,
                                    ICC_Function_Callback_t fnCallback, void *ptrUserData)
{
    int intReturnValue;
    assert(!sys_dnode_is_linked(&ptrFunctionCall->strNode)); // Maybe just return error?
    ptrFunctionCall->ptr_ICCContext = ptrContext;

    ICC_Packet_t *ptrPacket = icc_AllocateTXPacket(ptrContext, u32DataLength + sizeof(icc_FunctionHeader_t));
    if (ptrPacket != NULL)
    {
        ptrPacket->ptrSource             = NULL;
        ptrContext->u8_Transaction       = (ptrContext->u8_Transaction + 1) % 4;
        icc_FunctionHeader_t strFnHeader = {.strHeader.u3Type        = icc_u8PACKET_TYPE_FUNCTION,
                                            .strHeader.u2Transaction = ptrContext->u8_Transaction,
                                            .strHeader.flgEncrypted  = 0,
                                            .u32FunctionId           = sys_cpu_to_le32(u32FunctionId)};
        memcpy(ptrPacket->arrBuffer, &strFnHeader, sizeof(icc_FunctionHeader_t));
        if (ptrData != NULL)
        {
            memcpy(&ptrPacket->arrBuffer[sizeof(icc_FunctionHeader_t)], ptrData, u32DataLength);
            ptrPacket->u16Length = u32DataLength + sizeof(icc_FunctionHeader_t);
        }
        else
        {
            ptrPacket->u16Length = sizeof(icc_FunctionHeader_t);
        }

        ptrFunctionCall->fnCallback     = fnCallback;
        ptrFunctionCall->ptrUserData    = ptrUserData;
        ptrFunctionCall->u32FunctionId  = u32FunctionId;
        /** @todo Length is included in the CRC so it is the same as the CRC used in stream transactions. This is so
         * that the calculation can be consolidated into one place at some point.*/
        StackScop uint16_t u16_TempCRC  = crc16_ccitt(0xFFFF, (const uint8_t *)&ptrPacket->u16Length, sizeof(uint16_t));
        ptrFunctionCall->u16_RequestCRC = crc16_ccitt(u16_TempCRC, ptrPacket->arrBuffer, ptrPacket->u16Length);

        RTOS_MutexLock(&ptrContext->strHandlerLock, RTOS_FOREVER);
        sys_dlist_append(&ptrContext->strFunctionCalls, &ptrFunctionCall->strNode);
        RTOS_MutexUnlock(&ptrContext->strHandlerLock);
        icc_EnqueuePacket(ptrContext, ptrPacket);
        intReturnValue = 0;
    }
    else
    {
        intReturnValue = -ENOMEM;
    }
    return intReturnValue;
}

GlobScop void ICC_Function_InitCall(ICC_FunctionCall_t *ptrFunctionCall)
{
    ptrFunctionCall->ptr_ICCContext = NULL;
    sys_dnode_init(&ptrFunctionCall->strNode);
}

GlobScop void ICC_Function_DestroyCall(ICC_FunctionCall_t *ptrFunctionCall)
{
    if (ptrFunctionCall->ptr_ICCContext != NULL)
    {
        RTOS_MutexLock(&ptrFunctionCall->ptr_ICCContext->strHandlerLock, RTOS_FOREVER);
        if (sys_dnode_is_linked(&ptrFunctionCall->strNode))
        {
            sys_dlist_remove(&ptrFunctionCall->strNode);
        }
        RTOS_MutexUnlock(&ptrFunctionCall->ptr_ICCContext->strHandlerLock);
    }
}

/**********************************************************************************************************************
 * @cond                                                DLL SCOPE                                              @endcond
 *********************************************************************************************************************/

/* These functions will only ever be exposed when compiled into a DLL library as global symbols. */
#ifndef __ZEPHYR__
ModScop ICC_FunctionCall_t *ICC_Function_AllocateCall() { return malloc(sizeof(ICC_FunctionCall_t)); }
ModScop void                ICC_Function_FreeCall(ICC_FunctionCall_t *ptr_Context) { free(ptr_Context); }

ModScop ICC_Function_Handler_t *ICC_Function_AllocateHandler() { return malloc(sizeof(ICC_Function_Handler_t)); }
ModScop void                    ICC_Function_FreeHandler(ICC_Function_Handler_t *ptr_Context) { free(ptr_Context); }
#endif

/**********************************************************************************************************************
 * @cond                                              MODULE SCOPE                                             @endcond
 *********************************************************************************************************************/

ModScop bool icc_Function_ProcessReturnPacket(ICC_Context_t *ptrContext, ICC_Packet_t *ptrRspPacket)
{
    StackScop icc_ReturnHeader_t strRspHeader;
    if (ptrRspPacket->u16Length >= sizeof(icc_ReturnHeader_t))
    {
        memcpy(&strRspHeader, ptrRspPacket->arrBuffer, sizeof(icc_ReturnHeader_t));
        if (strRspHeader.strHeader.u3Type == icc_u8PACKET_TYPE_RETURN)
        {
            strRspHeader.u32FunctionId           = sys_le32_to_cpu(strRspHeader.u32FunctionId);
            strRspHeader.i32ReturnValue          = sys_le32_to_cpu(strRspHeader.i32ReturnValue);
            StackScop uint8_t *ptr_Payload       = &ptrRspPacket->arrBuffer[sizeof(icc_ReturnHeader_t)];
            StackScop uint16_t u16_PayloadLength = ptrRspPacket->u16Length - sizeof(icc_ReturnHeader_t);

            LOG_WRN("%p: Function Rsp Packet (%p) for 0x%08X[%d] = %d with %u bytes", ptrContext, ptrRspPacket,
                    strRspHeader.u32FunctionId, strRspHeader.strHeader.u2Transaction, strRspHeader.i32ReturnValue,
                    u16_PayloadLength);
            LOG_HEXDUMP_DBG(ptr_Payload, u16_PayloadLength, "Function Rsp Payload");

            StackScop ICC_FunctionCall_t *ptrHandler;
            StackScop ICC_FunctionCall_t *ptrNext;
            RTOS_MutexLock(&ptrContext->strHandlerLock, RTOS_FOREVER);
            SYS_DLIST_FOR_EACH_CONTAINER_SAFE(&ptrContext->strFunctionCalls, ptrHandler, ptrNext, strNode)
            {
                if (ptrHandler->u32FunctionId != strRspHeader.u32FunctionId)
                {
                    continue;
                }
                sys_dlist_remove(&ptrHandler->strNode);
                RTOS_MutexUnlock(&ptrContext->strHandlerLock);
                ptrHandler->fnCallback(strRspHeader.u32FunctionId, strRspHeader.i32ReturnValue, ptr_Payload,
                                       u16_PayloadLength, ptrHandler->ptrUserData);
                return true;
            }
            RTOS_MutexUnlock(&ptrContext->strHandlerLock);
        }
    }

    /** I really, REALLY don't like duplicating this code. But it was the cleanest way to handle it. */
    StackScop icc_ReturnHeaderV2_t strRspHeaderV2;
    if (ptrRspPacket->u16Length >= sizeof(icc_ReturnHeaderV2_t))
    {
        memcpy(&strRspHeaderV2, ptrRspPacket->arrBuffer, sizeof(icc_ReturnHeaderV2_t));
        if (strRspHeaderV2.strHeader.u3Type == icc_u8PACKET_TYPE_RETURNV2)
        {
            strRspHeaderV2.u32FunctionId         = sys_le32_to_cpu(strRspHeaderV2.u32FunctionId);
            strRspHeaderV2.i32ReturnValue        = sys_le32_to_cpu(strRspHeaderV2.i32ReturnValue);
            StackScop uint8_t *ptr_Payload       = &ptrRspPacket->arrBuffer[sizeof(icc_ReturnHeaderV2_t)];
            StackScop uint16_t u16_PayloadLength = ptrRspPacket->u16Length - sizeof(icc_ReturnHeaderV2_t);

            LOG_INF("%p: Function Rsp V2 Packet (%p) for 0x%08X[%d][%04X] = %d with %u bytes", ptrContext, ptrRspPacket,
                    strRspHeaderV2.u32FunctionId, strRspHeaderV2.strHeader.u2Transaction, strRspHeaderV2.u16_RequestCRC,
                    strRspHeaderV2.i32ReturnValue, u16_PayloadLength);
            LOG_HEXDUMP_DBG(ptr_Payload, u16_PayloadLength, "Function Rsp V2 Payload");

            StackScop ICC_FunctionCall_t *ptrHandler;
            StackScop ICC_FunctionCall_t *ptrNext;
            RTOS_MutexLock(&ptrContext->strHandlerLock, RTOS_FOREVER);
            SYS_DLIST_FOR_EACH_CONTAINER_SAFE(&ptrContext->strFunctionCalls, ptrHandler, ptrNext, strNode)
            {
                if ((ptrHandler->u32FunctionId != strRspHeaderV2.u32FunctionId) ||
                    (ptrHandler->u16_RequestCRC != strRspHeaderV2.u16_RequestCRC))
                {
                    continue;
                }
                sys_dlist_remove(&ptrHandler->strNode);
                RTOS_MutexUnlock(&ptrContext->strHandlerLock);
                ptrHandler->fnCallback(strRspHeaderV2.u32FunctionId, strRspHeaderV2.i32ReturnValue, ptr_Payload,
                                       u16_PayloadLength, ptrHandler->ptrUserData);
                return true;
            }
            RTOS_MutexUnlock(&ptrContext->strHandlerLock);
        }
    }
    return false;
}

ModScop bool icc_Function_ProcessFunctionPacket(ICC_Context_t *ptrContext, ICC_Packet_t *ptrReqPacket)
{
    StackScop icc_FunctionHeader_t strReqHeader;

    if (ptrReqPacket->u16Length >= sizeof(icc_FunctionHeader_t))
    {
        memcpy(&strReqHeader, ptrReqPacket->arrBuffer, sizeof(icc_FunctionHeader_t));
        if (strReqHeader.strHeader.u3Type == icc_u8PACKET_TYPE_FUNCTION)
        {
            strReqHeader.u32FunctionId           = sys_le32_to_cpu(strReqHeader.u32FunctionId);
            StackScop uint8_t *ptr_Payload       = &ptrReqPacket->arrBuffer[sizeof(icc_FunctionHeader_t)];
            StackScop uint16_t u16_PayloadLength = ptrReqPacket->u16Length - sizeof(icc_FunctionHeader_t);
            StackScop uint16_t u16_RequestCRC =
                crc16_ccitt(0xFFFF, (const uint8_t *)&ptrReqPacket->u16Length, sizeof(uint16_t));
            u16_RequestCRC = crc16_ccitt(u16_RequestCRC, ptrReqPacket->arrBuffer, ptrReqPacket->u16Length);

            LOG_INF("%p: Function Call Packet (%p) for 0x%08X[%d][%04X] with %u bytes", ptrContext, ptrReqPacket,
                    strReqHeader.u32FunctionId, strReqHeader.strHeader.u2Transaction, u16_RequestCRC,
                    u16_PayloadLength);
            LOG_HEXDUMP_DBG(ptr_Payload, u16_PayloadLength, "Function Call Payload");

            StackScop ICC_Function_Handler_t *ptrHandler;
            StackScop ICC_Function_Handler_t *ptrNext;
            RTOS_MutexLock(&ptrContext->strHandlerLock, RTOS_FOREVER);
            SYS_DLIST_FOR_EACH_CONTAINER_SAFE(&ptrContext->functions, ptrHandler, ptrNext, node)
            {
                if (ptrHandler->u32FunctionId != (ptrHandler->u32FunctionIdMask & strReqHeader.u32FunctionId))
                {
                    continue;
                }
                RTOS_MutexUnlock(&ptrContext->strHandlerLock);

                StackScop ICC_Packet_t *ptrResponse = icc_AllocateTXPacket(ptrContext, sizeof(icc_ReturnHeaderV2_t));
                if (ptrResponse != NULL)
                {
                    ptrResponse->ptrSource =
                        NULL; // TODO: Ideally, a function response should ONLY go to the source it was requested from.
                    StackScop uint32_t u32ResponseLength = ptrResponse->u16Length - sizeof(icc_ReturnHeaderV2_t);
                    StackScop int32_t  intResponseCode =
                        ptrHandler->fnCallback(strReqHeader.u32FunctionId, ptr_Payload, u16_PayloadLength,
                                               &ptrResponse->arrBuffer[sizeof(icc_ReturnHeaderV2_t)],
                                               &u32ResponseLength, ptrHandler->ptrUserData);
                    assert(u32ResponseLength <= (ptrResponse->u16Length - sizeof(icc_ReturnHeaderV2_t)));

                    /** @todo Length is included in the CRC so it is the same as the CRC used in stream transactions.
                     * This is so that the calculation can be consolidated into one place at some point.*/
                    StackScop icc_ReturnHeaderV2_t strFnHeader = {
                        .strHeader.u3Type        = icc_u8PACKET_TYPE_RETURNV2,
                        .strHeader.flgEncrypted  = 0,
                        .strHeader.u2Transaction = strReqHeader.strHeader.u2Transaction,
                        .u32FunctionId           = sys_cpu_to_le32(strReqHeader.u32FunctionId),
                        .i32ReturnValue          = sys_cpu_to_le32(intResponseCode),
                        .u16_RequestCRC          = u16_RequestCRC};
                    memcpy(ptrResponse->arrBuffer, &strFnHeader, sizeof(icc_ReturnHeaderV2_t));
                    ptrResponse->u16Length = u32ResponseLength + sizeof(icc_ReturnHeaderV2_t);

                    ptrResponse->ptrTarget = ptrReqPacket->ptrSource;
                    icc_EnqueuePacket(ptrContext, ptrResponse);
                }
                else
                {
                    LOG_DBG("%p: Unable to allocate response packet for call to 0x%08X", ptrContext,
                            strReqHeader.u32FunctionId);
                }

                return true;
            }
            RTOS_MutexUnlock(&ptrContext->strHandlerLock);
        }
    }
    return false;
}

/**********************************************************************************************************************
 * @cond                                               FILE SCOPE                                              @endcond
 *********************************************************************************************************************/

FileScop void icc_Function_SyncCallback(uint32_t u32FunctionId, int intResult, const void *ptrResponse,
                                        uint32_t u32ResponseLength, void *ptrUserData)
{
    StackScop icc_Function_SyncCallback_t *ptrSyncCallback = ptrUserData;

    ptrSyncCallback->i32ReturnValue = intResult;
    if (ptrSyncCallback->ptrResponse != NULL)
    {
        if (u32ResponseLength <= *(ptrSyncCallback->u32ResponseLength))
        {
            memcpy(ptrSyncCallback->ptrResponse, ptrResponse, u32ResponseLength);
            *(ptrSyncCallback->u32ResponseLength) = u32ResponseLength;
        }
        else
        {
            LOG_ERR("Response for 0x%08X was %u bytes, only given %u", u32FunctionId, u32ResponseLength,
                    *ptrSyncCallback->u32ResponseLength);
            ptrSyncCallback->i32ReturnValue = -ENOMEM;
        }
    }
    else if (ptrSyncCallback->u32ResponseLength != NULL)
    {
        *(ptrSyncCallback->u32ResponseLength) = u32ResponseLength;
    }

    RTOS_SemGive(&ptrSyncCallback->strSem);
}