#include "lion/log/log.h"
#include <ctype.h>
#include <pthread.h>
#include <stdbool.h>
#include <time.h>
/* cppcheck-suppress misra-c2012-21.6 */
#include <stdarg.h>
#include <stdio.h>

static pthread_mutex_t log_strMutex;
static bool            log_bolInitialized = false;

const char *log_LogLevelAsString(LOG_Level_t level)
{
    const char *result;
    switch (level)
    {
        case LOG_LEVEL_ERR:
            result = "err"; 
            break;
        case LOG_LEVEL_WRN:
            result = "wrn";
            break;
        case LOG_LEVEL_INF:
            result = "inf";
            break;
        case LOG_LEVEL_DBG:
            result = "dbg";
            break;
        case LOG_LEVEL_NONE:
        default:
            result = "none";
            break;
    }
    return result;
}

static inline void log_Divide(uint64_t numerator, uint64_t divisor, uint64_t *result, uint64_t *remainder)
{
    *result    = numerator / divisor;
    *remainder = numerator % divisor;
}

static void log_Prefix(LOG_Level_t level, const char *prefix)
{
    struct timespec strTime;
    clock_gettime(CLOCK_MONOTONIC, &strTime);
    uint64_t u64Hours;
    uint64_t u64Minutes;
    uint64_t u64Seconds;
    uint64_t u64Milliseconds;
    uint64_t u64Microseconds;
    log_Divide(strTime.tv_sec, 60, &u64Minutes, &u64Seconds);
    log_Divide(u64Minutes, 60, &u64Hours, &u64Minutes);
    u64Microseconds = strTime.tv_nsec / 1000;
    log_Divide(u64Microseconds, 1000, &u64Milliseconds, &u64Microseconds);

    printf("[%02u:%02u:%02u.%03u,%03u] <%s> %s: ", (uint32_t)u64Hours, (uint32_t)u64Minutes, (uint32_t)u64Seconds,
           (uint32_t)u64Milliseconds, (uint32_t)u64Microseconds, log_LogLevelAsString(level), prefix);
}

void log_DoLog(LOG_Level_t level, const char *prefix, const char *format, ...)
{
    if (!log_bolInitialized)
    {
        pthread_mutex_init(&log_strMutex, NULL);
        log_bolInitialized = true;
    }

    pthread_mutex_lock(&log_strMutex);

    log_Prefix(level, prefix);

    /* cppcheck-suppress misra-c2012-17.1 */
    va_list args;
    /* cppcheck-suppress misra-c2012-17.1 */
    va_start(args, format);
    vprintf(format, args);
    /* cppcheck-suppress misra-c2012-17.1 */
    va_end(args);

    printf("\n");
    
    fflush(stdout);

    pthread_mutex_unlock(&log_strMutex);
}

void log_DoHexDumpLog(LOG_Level_t level, const char *prefix, const void *data, uint32_t dataLength, const char *message)
{
    int            retval;
    char           szPrinted[17];
    const uint8_t *ptrBytes = data;
    uint32_t       u32Index = 0;

    if (!log_bolInitialized)
    {
        pthread_mutex_init(&log_strMutex, NULL);
        log_bolInitialized = true;
    }

    pthread_mutex_lock(&log_strMutex);

    log_Prefix(level, prefix);
    printf("%s\n", message);

    while (u32Index < dataLength)
    {
        printf("\t\t");

        for (uint8_t i = 0U; i < 16U; i++)
        {
            if ((u32Index + i) < dataLength)
            {
                printf("%02X ", ptrBytes[u32Index + i]);
            }
            else
            {
                printf("   ");
            }
        }

        printf("|");

        for (uint8_t i = 0U; i < 16U; i++)
        {
            char byte = ((u32Index + i) < dataLength) ? ptrBytes[u32Index + i] : ' ';
            printf("%c", isprint(byte) ? byte : '.');
        }

        printf("|\n");
        u32Index += 16;
    }

    printf("\n");

    fflush(stdout);

    pthread_mutex_unlock(&log_strMutex);
}