#include "lion/ICC_Client.h"
#include "lion/log/log.h"
#include "lion/rtos/RTOS_Sleep.h"
#include <errno.h>
#include <malloc.h>
#include <pthread.h>
#include <signal.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/un.h>

LOG_MODULE_DECLARE(icc_client);

FileScop ICC_Context_t str_IccCtx;
FileScop ICC_Client_Context_t str_IccClientCtx = {0};
FileScop pthread_t str_IccProcessThread;

/*********************************************************************************************************************
* @cond                                               FILE SCOPE                                              @endcond
*********************************************************************************************************************/

#define icc_client_sz_DEFAULT_SOCKET_PATH "/var/run/lion-icc.sock"
#define icc_client_u16_SLEEP_MS 10

FileScop void icc_client_SetState(ICC_Client_Context_t *ptr_str_Context, ICC_Client_State_t enm_State)
{
    RTOS_MutexLock(&ptr_str_Context->str_StateLock, RTOS_FOREVER);
    ptr_str_Context->enm_State = enm_State;
    RTOS_MutexUnlock(&ptr_str_Context->str_StateLock);
    RTOS_CondVarBroadcast(&ptr_str_Context->str_StateChanged);
}

FileScop size_t icc_client_StreamWrite(const uint8_t *ptr_u8_Data, size_t siz_Length, bool bol_All, void *vptr_UserData)
{
    ICC_Client_Context_t *ptr_str_Context = vptr_UserData;
    ssize_t ssiz_Result  = send(ptr_str_Context->fd_Socket, ptr_u8_Data, siz_Length, MSG_DONTWAIT | MSG_NOSIGNAL);
    if (ssiz_Result < 0)
    {
        ssiz_Result = 0;
        switch (errno)
        {
            case EAGAIN:
                /* Do Nothing */
                break;
            default:
                LOG_ERR("Unhandled Recv Error: \"%s\"", strerror(errno));
                ptr_str_Context->bol_Broken = true;
                break;
        }
    }
    return ssiz_Result;
}

FileScop size_t icc_client_StreamRead(uint8_t *ptr_u8_Data, size_t siz_MaxLength, bool bol_All, void *vptr_UserData)
{
    StackScop ICC_Client_Context_t *ptr_str_Context = vptr_UserData;
    StackScop ssize_t ssiz_Result = recv(ptr_str_Context->fd_Socket, ptr_u8_Data, siz_MaxLength, 
                                            MSG_DONTWAIT | (bol_All ? MSG_WAITALL : 0));
    if (ssiz_Result < 0)
    {
        ssiz_Result = 0;
        switch (errno)
        {
            case EAGAIN:
                /* Do Nothing */
                break;
            default:
                LOG_ERR("Unhandled Recv Error: \"%s\"", strerror(errno));
                ptr_str_Context->bol_Broken = true;
                break;
        }
    }
    return ssiz_Result;
}

FileScop void *icc_client_ProcessThreadEntry(void *arg)
{
    (void)arg;
    while (str_IccClientCtx.bol_KeepRunning)
    {
        ICC_Process(str_IccClientCtx.ptr_ICC);
        RTOS_Sleep(RTOS_MSEC(icc_client_u16_SLEEP_MS));
    }
}

FileScop void *icc_client_ThreadEntry(void *arg)
{
    StackScop ICC_Client_Context_t *ptr_str_Context = arg;

    while (ptr_str_Context->bol_KeepRunning)
    {
        icc_client_SetState(ptr_str_Context, ICC_CLIENT_DISCONNECTED);
        ptr_str_Context->fd_Socket = socket(AF_UNIX, SOCK_STREAM, 0);
        if (ptr_str_Context->fd_Socket < 0)
        {
            LOG_ERR("Unable to create socket \"%s\"", strerror(errno));
            RTOS_Sleep(RTOS_SECONDS(1));
            continue;
        }

        StackScop struct sockaddr_un str_SocketAddress;
        str_SocketAddress.sun_family = AF_UNIX;
        strcpy(str_SocketAddress.sun_path, ptr_str_Context->sz_SocketPath);

        if (connect(ptr_str_Context->fd_Socket, (const struct sockaddr *)&str_SocketAddress, sizeof(str_SocketAddress)))
        {
            LOG_ERR("Unable to connect socket \"%s\"", strerror(errno));
            RTOS_Sleep(RTOS_SECONDS(1));
            shutdown(ptr_str_Context->fd_Socket, SHUT_RDWR);
            continue;
        }

        ptr_str_Context->bol_Broken = false;
        ICC_Stream_Init(&ptr_str_Context->str_Stream, icc_client_StreamWrite, icc_client_StreamRead, ptr_str_Context);
        ICC_AddTransport(ptr_str_Context->ptr_ICC, &ptr_str_Context->str_Stream.transport);

        LOG_INF("Connected");
        icc_client_SetState(ptr_str_Context, ICC_CLIENT_CONNECTED);

        while (ptr_str_Context->bol_KeepRunning && !ptr_str_Context->bol_Broken)
        {
            RTOS_Sleep(RTOS_MSEC(100));
        }
        ICC_RemoveTransport(ptr_str_Context->ptr_ICC, &ptr_str_Context->str_Stream.transport);

        LOG_INF("Disconnected");

        shutdown(ptr_str_Context->fd_Socket, SHUT_RDWR);
    }

    icc_client_SetState(ptr_str_Context, ICC_CLIENT_DISABLED);

    return NULL;
}

/*********************************************************************************************************************
* @cond                                              GLOBAL SCOPE                                             @endcond
*********************************************************************************************************************/

GlobScop ICC_Context_t *ICC_Client_Init(const char *sz_SocketPath)
{
    str_IccClientCtx.ptr_ICC         = &str_IccCtx;
    ICC_Init(str_IccClientCtx.ptr_ICC);
    str_IccClientCtx.sz_SocketPath   = (sz_SocketPath == NULL) ? icc_client_sz_DEFAULT_SOCKET_PATH : sz_SocketPath;
    str_IccClientCtx.bol_KeepRunning = false;
    str_IccClientCtx.enm_State       = ICC_CLIENT_DISABLED;
    RTOS_MutexInit(&str_IccClientCtx.str_StateLock);
    RTOS_CondVarInit(&str_IccClientCtx.str_StateChanged);
    return str_IccClientCtx.ptr_ICC;
}

GlobScop int  ICC_Client_Start(RTOS_Timeout_t tim_Timeout)
{
    StackScop int int_Result = 0;
    RTOS_MutexLock(&str_IccClientCtx.str_StateLock, RTOS_FOREVER);
    if (str_IccClientCtx.bol_KeepRunning)
    {
        /* Already Running: Do nothing */
        int_Result = -EALREADY;
    }
    else
    {
        str_IccClientCtx.bol_KeepRunning = true;
        pthread_create(&str_IccProcessThread, NULL, icc_client_ProcessThreadEntry, NULL);
        pthread_create(&str_IccClientCtx.fd_Thread, NULL, icc_client_ThreadEntry, &str_IccClientCtx);
        ICC_Client_WaitForConnected(&str_IccClientCtx, tim_Timeout);
    }
    RTOS_MutexUnlock(&str_IccClientCtx.str_StateLock);

    return int_Result;
}

GlobScop int ICC_Client_Stop(void)
{
    StackScop int int_Result = 0;
    RTOS_MutexLock(&str_IccClientCtx.str_StateLock, RTOS_FOREVER);
    if (str_IccClientCtx.bol_KeepRunning)
    {
        str_IccClientCtx.bol_KeepRunning = false;
        /* TODO: Use pthread_kill() to interrupt any system calls */
        // pthread_kill(str_IccClientCtx.fd_Thread, SIGUSR1);
        pthread_join(str_IccProcessThread, NULL);
        pthread_join(str_IccClientCtx.fd_Thread, NULL);
    }
    else
    {
        /* Already Stopped: Do nothing */
        int_Result = -EALREADY;
    }
    RTOS_MutexUnlock(&str_IccClientCtx.str_StateLock);

    return int_Result;
}

GlobScop ICC_Client_State_t ICC_Client_GetState(void)
{
    RTOS_MutexLock(&str_IccClientCtx.str_StateLock, RTOS_FOREVER);
    ICC_Client_State_t enm_State = str_IccClientCtx.enm_State;
    RTOS_MutexUnlock(&str_IccClientCtx.str_StateLock);
    return enm_State;
}

GlobScop int ICC_Client_WaitForState(ICC_Client_Context_t *ptr_str_Context, ICC_Client_State_t enm_State, 
                                        RTOS_Timeout_t tim_Timeout)
{
    StackScop int int_RC = 0;
    RTOS_Timepoint_t tim_Timepoint = RTOS_TimepointCalc(tim_Timeout);
    RTOS_MutexLock(&ptr_str_Context->str_StateLock, RTOS_FOREVER);
    while (ptr_str_Context->enm_State != enm_State)
    {
        int_RC = RTOS_CondVarWait(&ptr_str_Context->str_StateChanged, &ptr_str_Context->str_StateLock,
                                    RTOS_TimepointToTimeout(tim_Timepoint));
        if (int_RC == -EAGAIN)
        {
            break;
        }
    }
    RTOS_MutexUnlock(&ptr_str_Context->str_StateLock);
    return int_RC;
}