/**********************************************************************************************************************
 * @file            LPP.h
 *
 * @brief           The Lion Protocol Processor (LPP) service provides a common interface for accepting incoming
 *                  messages from registered transports, and passing them to function handlers to be processed.
 *
 * @author          Abram Early
 * @date            08/08/2023
 * @copyright       2023 Lion Energy LLC.
 *                  All rights reserved
 *
 *********************************************************************************************************************/
/**
 * @addtogroup      LPP
 *                  Lion Protocol Processor (LPP) Service Module
 *
 * @ref             LPP.h
 *
 *********************************************************************************************************************/
#include "COD_Compiler_Defines.h"
#include "lion/rtos/RTOS_Timeout.h"
#include "lion/utils/dlist.h"

#ifndef LION_LPP_H__
    #define LION_LPP_H__

    #ifdef __cplusplus
        #include <ArduinoJson.h>

typedef struct LPP_Transport_t LPP_Transport_t;

typedef struct LPP_TXContext_t
{
    LPP_Transport_t *ptr_Transport;
    void            *ptr_Destination;

    uint16_t u16_Index;
    uint16_t u16_Count;
    size_t   siz_MTU;
    size_t   siz_Offset;
    size_t   siz_Length;
} LPP_TXContext_t;

typedef struct LPP_Request_t
{
    const char *sz_Id;
    const char *sz_Type;

    /**
     * @brief Reference to the body of the request.
     */
    const ArduinoJson::JsonObject &ref_Request;
    /**
     * @brief The transport the request originated from.
     */
    LPP_Transport_t               *ptr_Transport;
    /**
     * @brief Transport implementation specific source pointer. i.e. A pointer to the BLE connection definition.
     */
    void                          *ptr_Source;

    LPP_TXContext_t str_Response;
} LPP_Request_t;
    #else
/* This struct can only be defined in cpp files, but we can still use a pointer to it in C, so we just need a
 * placeholder. */
typedef struct LPP_Request_t LPP_Request_t;
typedef struct LPP_TXContext_t LPP_TXContext_t;
    #endif

    #ifdef __cplusplus
extern "C"
{
    #endif

    /**********************************************************************************************************************
     * @cond                                              GLOBAL SCOPE @endcond
     *********************************************************************************************************************/

    /**
     * @brief Called by LPP to handle a request.
     * @param ptr_Request Reference to the incoming request and its details.
     * @return 0 on Success, or any negative standard POSIX error.
     */
    typedef int (*LPP_HandlerCallback_t)(LPP_Request_t *ptr_Request);

    /**
     * @brief Called by LPP to send message via a transport.
     * @param ptr_Destination Arbitrary, transport defined pointer to identify the client. i.e. Pointer to WebSocket
     * client object
     * @param ptr_Data Pointer to the message to be sent.
     * @param siz_Length The length of the message to be sent
     * @param tim_Timeout A timeout at which the callback should cancel any operation, and return @c -EAGAIN
     * @param ptr_UserData Arbitrary, User-Defined pointer set during transport initialization
     */
    typedef int (*LPP_TransportTX_t)(void *ptr_Destination, const uint8_t *ptr_Data, size_t siz_Length,
                                     RTOS_Timeout_t tim_Timeout, void *ptr_UserData);

    /* The following are intended to be implemented, but the LPP doesn't support it yet.
        typedef ssize_t (*LPP_TransportRX_t)(void *ptr_Source, uint8_t *ptr_Data, size_t siz_Length, RTOS_Timeout_t
       tim_Timeout, void *ptr_UserData); typedef size_t (*LPP_TransportGetMTU_t)(void *ptr_Source, void *ptr_UserData);
        */

    typedef struct LPP_FunctionHandler_t
    {
        const char           *sz_Name;
        LPP_HandlerCallback_t fn_Get;
        LPP_HandlerCallback_t fn_Set;
    } LPP_FunctionHandler_t;

    typedef struct LPP_Transport_t
    {
        sys_dnode_t       str_Node;
        // LPP_TransportRX_t fn_RX;
        /**
         * @brief Called by LPP to send message via a transport.
         * @ref LPP_TransportTX_t
         */
        LPP_TransportTX_t fn_TX;
        // LPP_TransportGetMTU_t fn_GetMTU;
        void             *ptr_UserData;
    } LPP_Transport_t;

    typedef struct LPP_Handlers_t
    {
        sys_dnode_t            str_Node;
        LPP_FunctionHandler_t *ptr_HandlerList;
    } LPP_Handlers_t;

    /**
     * @brief Initialize the LPP module. This should only be called once.
     * @warning No other functions in the module should be called until LPP_Init is called.
     *
     * @param ptr_ICC Pointer to a working ICC context
     *
     */
    GlobScop void LPP_Init();

    /**
     * @brief Starts the LPP module
     *
     * @returns 0 on success,
     * @returns -EALREADY if already running.
     */
    GlobScop int LPP_Start();

    /**
     * @brief Terminates the LPP module
     * @param tim_Timeout Timeout to wait for module to shutdown, or RTOS_NO_WAIT to return immediately regardless of
     * success.
     *
     * @returns 0 on success or if tim_Timeout is RTOS_NO_WAIT
     * @returns -EAGAIN if timed out waiting for module to shutdown
     */
    GlobScop int LPP_Stop(RTOS_Timeout_t tim_Timeout);

    /**
     * @brief Initializes a LPP_Handlers_t structure
     * @param ptr_Handlers Pointer to LPP_Handlers_t structure
     * @param ptr_HandlerList Pointer to an array of LPP_FunctionHandler_t. The last item MUST contain a NULL pointer in
     * sz_Name.
     */
    GlobScop void LPP_InitHandlers(LPP_Handlers_t *ptr_Handlers, LPP_FunctionHandler_t *ptr_HandlerList);

    /**
     * @brief Registers the specified handlers
     * @param ptr_Handlers Pointer to LPP_Handlers_t structure
     * @returns 0 on success
     * @returns -EALREADY if already registered
     */
    GlobScop int LPP_RegisterHandlers(LPP_Handlers_t *ptr_Handlers);

    /**
     * @brief Unregisters the specified handlers
     * @param ptr_Handlers Pointer to LPP_Handlers_t structure
     * @returns 0 on success
     * @returns -EALREADY if already unregistered
     */
    GlobScop int LPP_UnregisterHandlers(LPP_Handlers_t *ptr_Handlers);

    /**
     * @brief Initializes a LPP_Transport_t structure
     * @param ptr_Transport Pointer to LPP_Transport_t structure
     * @param fn_TX Callback for transmitting data back to client
     * @param ptr_UserData Arbitrary, User-Defined pointer that is passed back to any callbacks.
     */
    GlobScop void LPP_InitTransport(LPP_Transport_t *ptr_Transport, LPP_TransportTX_t fn_TX, void *ptr_UserData);

    /**
     * @brief Registers the specified transport
     * @param ptr_Transport Pointer to LPP_Transport_t structure
     * @returns 0 on success
     * @returns -EALREADY if already registered
     */
    GlobScop int LPP_RegisterTransport(LPP_Transport_t *ptr_Transport);

    /**
     * @brief Unregisters the specified transport
     * @param ptr_Transport Pointer to LPP_Transport_t structure
     * @returns 0 on success
     * @returns -EALREADY if already unregistered
     */
    GlobScop int LPP_UnregisterTransport(LPP_Transport_t *ptr_Transport);

    /**
     * @brief Reads up to @p siz_Length bytes from the incoming request specified by @p ptr_Request
     * @param ptr_Request Reference to the request provided when a @c LPP_HandlerCallback_t function is called.
     * @param ptr_Destination Destination buffer to write bytes to
     * @param siz_Length Should be set to the number of bytes to read into @c ptr_Destination when calling. Will return
     * the number of read bytes regardless of return code.
     * @param tim_Timeout Time to wait for @p siz_Length bytes, or @c RTOS_NO_WAIT, or @c RTOS_FOREVER
     * @returns 0 On Success
     * @returns @c -EAGAIN If a timeout occured while waiting for data.
     * @returns @c -ENOENT If all bytes in the message were read.
     * @warning Must ONLY be used within @c LPP_HandlerCallback_t functions called internally by the LPP.
     */
    GlobScop int LPP_Receive(LPP_Request_t *ptr_Request, void *ptr_Destination, size_t *siz_Length,
                             RTOS_Timeout_t tim_Timeout);

    /**
     * @brief Submits a complete frame to the LPP processor.
     * @param ptr_Transport Pointer to the source @c LPP_Transport_t object.
     * @param ptr_Source Arbitrary, transport defined pointer that should identify the client. i.e. Pointer to WebSocket
     * client object
     * @param ptr_Data Frame data
     * @param siz_Length Length of the frame
     * @return 0 on Success
     * @return @c -EINVAL if frame is invalid
     * @return @c -EBUSY if another transport is being processed
     * @warning This is a temporary function that will be replaced by a more robust mechanism later.
     */
    GlobScop int LPP_RXFrame(LPP_Transport_t *ptr_Transport, void *ptr_Source, const uint8_t *ptr_Data,
                             size_t siz_Length);

    #ifdef __cplusplus
}

/**
 * @brief Submits a response to the incoming request specified by @p ptr_Request with @p ref_Response as the "body", and
 * specifies how many raw bytes will follow.
 * @param ptr_Request Reference to the request provided when a @c LPP_HandlerCallback_t function is called.
 * @param ref_Response The JSON document to return as the "body" of the response.
 * @param siz_ContentLength The number of bytes that will subsequently be sent via the @c LPP_ResponseSend function.
 * @param tim_Timeout Time to wait for @p siz_Length bytes, or @c RTOS_NO_WAIT, or @c RTOS_FOREVER
 * @returns 0 On Success
 * @returns @c -EAGAIN If a timeout occured while waiting to write data.
 * @warning Must ONLY be used within @c LPP_HandlerCallback_t functions called internally by the LPP.
 */
GlobScop int LPP_Respond(LPP_Request_t *ptr_Request, const ArduinoJson::JsonVariant &ref_Response,
                         size_t siz_ContentLength, RTOS_Timeout_t tim_Timeout);

/**
 * @brief Sends up to @p siz_Length bytes as a response to the incoming request specified by @p ptr_Request
 * @param ptr_Request Reference to the request provided when a @c LPP_HandlerCallback_t function is called.
 * @param ptr_Content Source buffer to read bytes from
 * @param siz_Length Should be set to the number of bytes to write from @c ptr_Content when calling. Will return the
 *                   number of written bytes regardless of return code.
 * @param tim_Timeout Time to wait for @p siz_Length bytes, or @c RTOS_NO_WAIT, or @c RTOS_FOREVER
 * @returns 0 On Success
 * @returns @c -EAGAIN If a timeout occured while waiting to write data.
 * @warning Must ONLY be used within @c LPP_HandlerCallback_t functions called internally by the LPP.
 */
GlobScop int LPP_ResponseSend(LPP_Request_t *ptr_Requst, const void *ptr_Content, size_t *siz_Length,
                              RTOS_Timeout_t tim_Timeout);

    #endif

#endif /* LION_LPP_H__ */