#include "lion/LPP.h"
#include "lion/log/log.h"
#include "lion/rtos/RTOS_Mutex.h"
#include "lion/rtos/RTOS_Sem.h"
#include "lion/rtos/RTOS_Sleep.h"
#include <cassert>

LOG_MODULE_REGISTER(lpp, LOG_LEVEL_DBG);

/**********************************************************************************************************************
 * @cond                                               FILE SCOPE                                              @endcond
 *********************************************************************************************************************/

FileScop pthread_t     lpp_str_Thread;
FileScop RTOS_Mutex_t  lpp_str_Mutex;
FileScop volatile bool lpp_bol_KeepRunning;
FileScop sys_dlist_t   lpp_str_HandlerList;
FileScop sys_dlist_t   lpp_str_TransportList;

FileScop ArduinoJson::StaticJsonDocument<8192> lpp_cls_RequestBuffer;
FileScop uint8_t                               lpp_arr_RawBuffer[8192];
FileScop RTOS_Sem_t                            lpp_str_ActiveReady;
FileScop RTOS_Mutex_t                          lpp_str_ActiveMutex;
FileScop LPP_Transport_t                      *lpp_ptr_ActiveTransport;
FileScop void                                 *lpp_ptr_ActiveSource;
FileScop uint32_t                              lpp_u32_ActiveSequence;
FileScop size_t                                lpp_siz_ActiveLength;
FileScop size_t                                lpp_siz_ActiveReadPosition;

/**
 * @brief Initialize a transmission to the designated transport and source.
 * @param ptr_Context Transmission Context to Initialize
 * @param ptr_Transport Pointer to the source @c LPP_Transport_t object.
 * @param ptr_Source Arbitrary, transport defined pointer that should identify the client. i.e. Pointer to WebSocket
 * client object
 * @param siz_Length Length of transmission to expect
 * @param tim_Timeout Time to wait for @p siz_Length bytes, or @c RTOS_NO_WAIT, or @c RTOS_FOREVER
 * @return
 */
FileScop int lpp_TXInit(LPP_TXContext_t *ptr_Context, LPP_Transport_t *ptr_Transport, void *ptr_Destination,
                        size_t siz_Length, RTOS_Timeout_t tim_Timeout)
{
    assert(ptr_Context != NULL);
    assert(ptr_Transport != NULL);

    ptr_Context->ptr_Transport   = ptr_Transport;
    ptr_Context->ptr_Destination = ptr_Destination;
    ptr_Context->siz_MTU         = sizeof(lpp_arr_RawBuffer); /** TODO: Get from Transport */
    ptr_Context->siz_Offset      = 0;
    ptr_Context->siz_Length      = siz_Length;
    ptr_Context->u16_Index       = 1;

    /**
     * @todo The u16_Count is utterly broken. It will not return the correct number of packets since the number
     * of bytes sendable per message changes depending on the index and count since they are encoded as text. WCM gets
     * around this by just cutting 10 bytes off the MTU and only sending that number of bytes after the sequence tag.
     * Though the rest of LPP_Response* is setup to handle MTU, I'm just bypassing this entirely by using an 8kb MTU
     * since Websockets don't technically have an MTU. AKA: I can't be bothered to properly fix it right now.
     */
    ptr_Context->u16_Count = (ptr_Context->siz_Length + ptr_Context->siz_MTU - 1) / ptr_Context->siz_MTU;

    LOG_DBG("TX %p for %p.%p: %d bytes, %d frames", ptr_Context, ptr_Transport, ptr_Destination, siz_Length,
            ptr_Context->u16_Count);
    return 0;
}

/**
 * @brief Sends the provided block of data to the designated TX context. The message is broken into MTU sized
 * chunks as needed. Passing NULL to both @p ptr_Source and @p siz_Length will enter "Termination Mode" and will submit
 * any pushed data, and then transmit empty frames as needed to finish the transaction.
 * @param ptr_Context  Transmission Context Pointer
 * @param ptr_Source Data Pointer
 * @param[inout] siz_Length Should be set to the number of bytes to write from @c ptr_Source when calling. Will return
 * the number of written bytes regardless of return code.
 * @param tim_Timeout Time to wait for transmission, or @c RTOS_NO_WAIT, or @c RTOS_FOREVER
 * @return 0 on Success
 */
FileScop int lpp_TXSend(LPP_TXContext_t *ptr_Context, const void *ptr_Source, size_t *siz_Length,
                        RTOS_Timeout_t tim_Timeout)
{
    StackScop bool bol_TerminationMode = (ptr_Source == NULL) && (siz_Length == NULL);
    assert(ptr_Context != NULL);
    assert(bol_TerminationMode || ((ptr_Source != NULL) && (siz_Length != NULL)));

    StackScop int      err_RC;
    StackScop uint8_t *ptr_Bytes = (uint8_t *)ptr_Source;
    while ((bol_TerminationMode && (ptr_Context->u16_Index <= ptr_Context->u16_Count)) ||
           ((siz_Length != NULL) && (*siz_Length > 0)))
    {
        if (ptr_Context->siz_Offset == 0)
        {
            /// @todo Take possession of TX buffer
            err_RC = sprintf((char *)lpp_arr_RawBuffer, "%u/%u:", ptr_Context->u16_Index, ptr_Context->u16_Count);
            if (err_RC < 0)
            {
                /// @todo Panic?
            }
            else
            {
                ptr_Context->siz_Offset += err_RC;
            }
        }
        if (!bol_TerminationMode)
        {
            StackScop size_t siz_RemainingOfMTU = ptr_Context->siz_MTU - ptr_Context->siz_Offset;
            StackScop size_t siz_ToSend         = siz_RemainingOfMTU > *siz_Length ? *siz_Length : siz_RemainingOfMTU;
            memcpy(&lpp_arr_RawBuffer[ptr_Context->siz_Offset], ptr_Bytes, siz_ToSend);
            ptr_Context->siz_Offset += siz_ToSend;
            *siz_Length -= siz_ToSend;
            ptr_Bytes = &ptr_Bytes[siz_ToSend];
        }
        if (bol_TerminationMode || (ptr_Context->siz_Offset >= ptr_Context->siz_MTU))
        {
            err_RC = ptr_Context->ptr_Transport->fn_TX(ptr_Context->ptr_Destination, lpp_arr_RawBuffer,
                                                       ptr_Context->siz_Offset, tim_Timeout,
                                                       ptr_Context->ptr_Transport->ptr_UserData);
            /// @todo Check error

            LOG_DBG("TX %p for %p.%p: frame %d/%d", ptr_Context, ptr_Context->ptr_Transport,
                    ptr_Context->ptr_Destination, ptr_Context->u16_Index, ptr_Context->u16_Count);

            ptr_Context->siz_Offset = 0;
            ptr_Context->u16_Index += 1;
            /// @todo Release possession of TX buffer
        }
    }
    return 0;
}

/**
 * @brief Uses "Termination Mode" of @c lpp_TXSend to finish the transaction.
 * @param tim_Timeout Time to wait for transmission, or @c RTOS_NO_WAIT, or @c RTOS_FOREVER
 * @return 0 on Success
 */
FileScop int lpp_TXEnd(LPP_TXContext_t *ptr_Context, RTOS_Timeout_t tim_Timeout)
{
    return lpp_TXSend(ptr_Context, NULL, NULL, tim_Timeout);
}

struct TXWriter
{
    LPP_TXContext_t *ptr_Context;
    RTOS_Timeout_t   tim_ABSTimeout;

  public:
    explicit TXWriter(LPP_TXContext_t *ptr_Context, RTOS_Timeout_t tim_Timeout)
        : ptr_Context(ptr_Context), tim_ABSTimeout(RTOS_TIMEOUT_ABS(tim_Timeout))
    {
    }
    size_t write(uint8_t u8_Value) { return write(&u8_Value, 1); }
    size_t write(const uint8_t *ptr_Data, size_t siz_Length)
    {
        size_t siz_WriteLength = siz_Length;
        (void)lpp_TXSend(ptr_Context, ptr_Data, &siz_WriteLength, tim_ABSTimeout);
        return siz_WriteLength;
    }
};

/**
 * @brief A helper struct to get the number of bytes processed as JSON.
 */
struct RawReader
{
    const char *sz_Pointer;
    const char *sz_Start;

  public:
    explicit RawReader(const void *ptr_Raw)
        : sz_Pointer(ptr_Raw ? reinterpret_cast<const char *>(ptr_Raw) : ""), sz_Start(sz_Pointer)
    {
    }
    int    read() { return static_cast<unsigned char>(*sz_Pointer++); }
    size_t readBytes(char *ptr_Bytes, size_t siz_Length)
    {
        for (size_t i = 0; i < siz_Length; i++)
        {
            ptr_Bytes[i] = *sz_Pointer++;
        }
        return siz_Length;
    }
    size_t totalRead() { return sz_Pointer - sz_Start; }
};

/**
 * @brief Processing thread entry for LPP
 * @param ptr_Arg Unused
 */
FileScop void *lpp_ThreadEntry(void *ptr_Arg)
{
    LOG_INF("Started");
    while (lpp_bol_KeepRunning)
    {
        if (RTOS_SemTake(&lpp_str_ActiveReady, RTOS_MSEC(100)) != 0)
        {
            RTOS_MutexLock(&lpp_str_ActiveMutex, RTOS_FOREVER);
            lpp_ptr_ActiveTransport = NULL;
            lpp_ptr_ActiveSource    = NULL;
            lpp_u32_ActiveSequence  = 0;
            lpp_siz_ActiveLength    = 0;
            RTOS_MutexUnlock(&lpp_str_ActiveMutex);
            continue;
        }

        StackScop RawReader cls_Reader(lpp_arr_RawBuffer);
        StackScop ArduinoJson::DeserializationError cls_Error =
            ArduinoJson::deserializeJson(lpp_cls_RequestBuffer, cls_Reader);
        lpp_siz_ActiveReadPosition = cls_Reader.totalRead();

        if (!(bool)cls_Error)
        {
            StackScop LPP_Request_t str_Request = {.sz_Id         = lpp_cls_RequestBuffer["id"] | (const char *)NULL,
                                                   .sz_Type       = lpp_cls_RequestBuffer["type"] | (const char *)NULL,
                                                   .ref_Request   = lpp_cls_RequestBuffer["body"],
                                                   .ptr_Transport = lpp_ptr_ActiveTransport,
                                                   .ptr_Source    = lpp_ptr_ActiveSource};

            if (str_Request.sz_Id != NULL && str_Request.sz_Type != NULL)
            {
                StackScop LPP_FunctionHandler_t *ptr_Callback = NULL;
                StackScop LPP_Handlers_t        *ptr_HandlerList;
                StackScop LPP_Handlers_t        *ptr_Next;
                SYS_DLIST_FOR_EACH_CONTAINER_SAFE(&lpp_str_HandlerList, ptr_HandlerList, ptr_Next, str_Node)
                {
                    ptr_Callback = ptr_HandlerList->ptr_HandlerList;
                    while (ptr_Callback->sz_Name != NULL)
                    {
                        if (strcmp(ptr_Callback->sz_Name, str_Request.sz_Id) == 0)
                        {
                            break;
                        }
                        else
                        {
                            ptr_Callback = &ptr_Callback[1];
                        }
                    }
                    if (ptr_Callback != NULL)
                    {
                        break;
                    }
                }

                StackScop LPP_HandlerCallback_t fn_Callback = NULL;
                if (ptr_Callback != NULL)
                {
                    if (strcmp(str_Request.sz_Type, "get") == 0)
                    {
                        fn_Callback = ptr_Callback->fn_Get;
                    }
                    else if (strcmp(str_Request.sz_Type, "set") == 0)
                    {
                        fn_Callback = ptr_Callback->fn_Set;
                    }
                }

                StackScop ArduinoJson::StaticJsonDocument<96> cls_ResponseBuffer;
                cls_ResponseBuffer.clear();
                cls_ResponseBuffer["id"]   = str_Request.sz_Id;
                cls_ResponseBuffer["type"] = str_Request.sz_Type;

                if (fn_Callback != NULL)
                {
                    cls_ResponseBuffer["st"] = "ack";

                    StackScop LPP_TXContext_t str_TXContext;
                    (void)lpp_TXInit(&str_TXContext, str_Request.ptr_Transport, str_Request.ptr_Source,
                                     measureJson(cls_ResponseBuffer), RTOS_SECONDS(3));
                    TXWriter cls_TXWriter(&str_TXContext, RTOS_SECONDS(3));
                    serializeJson(cls_ResponseBuffer, cls_TXWriter);
                    (void)lpp_TXEnd(&str_TXContext, RTOS_SECONDS(3));

                    str_Request.str_Response.ptr_Transport = NULL;

                    (void)fn_Callback(&str_Request);

                    if (str_Request.str_Response.ptr_Transport != NULL)
                    {
                        (void)lpp_TXEnd(&str_Request.str_Response, RTOS_SECONDS(3));
                    }
                }
                else
                {
                    cls_ResponseBuffer["st"] = "inval";

                    StackScop LPP_TXContext_t str_TXContext;
                    (void)lpp_TXInit(&str_TXContext, str_Request.ptr_Transport, str_Request.ptr_Source,
                                     measureJson(cls_ResponseBuffer), RTOS_SECONDS(3));
                    TXWriter cls_TXWriter(&str_TXContext, RTOS_SECONDS(3));
                    serializeJson(cls_ResponseBuffer, cls_TXWriter);
                    (void)lpp_TXEnd(&str_TXContext, RTOS_SECONDS(3));
                }
            }
        }

        RTOS_MutexLock(&lpp_str_ActiveMutex, RTOS_FOREVER);
        lpp_ptr_ActiveTransport = NULL;
        lpp_ptr_ActiveSource    = NULL;
        RTOS_MutexUnlock(&lpp_str_ActiveMutex);
    }
    LOG_INF("Terminated");

    return NULL;
}

/**********************************************************************************************************************
 * @cond                                              GLOBAL SCOPE                                             @endcond
 *********************************************************************************************************************/

GlobScop void LPP_Init()
{
    RTOS_MutexInit(&lpp_str_Mutex);
    RTOS_MutexInit(&lpp_str_ActiveMutex);
    RTOS_SemInit(&lpp_str_ActiveReady, 0, 1);
    lpp_bol_KeepRunning = false;
    sys_dlist_init(&lpp_str_HandlerList);
    sys_dlist_init(&lpp_str_TransportList);
}

GlobScop int LPP_Start()
{
    int err_RC;
    RTOS_MutexLock(&lpp_str_Mutex, RTOS_FOREVER);
    if (lpp_bol_KeepRunning)
    {
        err_RC = -EALREADY;
    }
    else
    {
        lpp_bol_KeepRunning = true;
        pthread_create(&lpp_str_Thread, NULL, lpp_ThreadEntry, NULL);
        err_RC = 0;
    }
    RTOS_MutexUnlock(&lpp_str_Mutex);
    return err_RC;
}

GlobScop int LPP_Stop(RTOS_Timeout_t tim_Timeout)
{
    int err_RC;
    RTOS_MutexLock(&lpp_str_Mutex, RTOS_FOREVER);
    if (lpp_bol_KeepRunning)
    {
        lpp_bol_KeepRunning = false;
        pthread_join(lpp_str_Thread, NULL);
        err_RC = 0;
    }
    else
    {
        err_RC = -EALREADY;
    }
    RTOS_MutexUnlock(&lpp_str_Mutex);
    return err_RC;
}

GlobScop void LPP_InitHandlers(LPP_Handlers_t *ptr_Handlers, LPP_FunctionHandler_t *ptr_HandlerList)
{
    sys_dnode_init(&ptr_Handlers->str_Node);
    ptr_Handlers->ptr_HandlerList = ptr_HandlerList;
}

GlobScop int LPP_RegisterHandlers(LPP_Handlers_t *ptr_Handlers)
{
    if (sys_dnode_is_linked(&ptr_Handlers->str_Node))
    {
        return -EALREADY;
    }
    else
    {
        sys_dlist_append(&lpp_str_HandlerList, &ptr_Handlers->str_Node);
        return 0;
    }
}

GlobScop int LPP_UnregisterHandlers(LPP_Handlers_t *ptr_Handlers)
{
    if (sys_dnode_is_linked(&ptr_Handlers->str_Node))
    {
        sys_dlist_remove(&ptr_Handlers->str_Node);
        return 0;
    }
    else
    {
        return -EALREADY;
    }
}

GlobScop void LPP_InitTransport(LPP_Transport_t *ptr_Transport, LPP_TransportTX_t fn_TX, void *ptr_UserData)
{
    sys_dnode_init(&ptr_Transport->str_Node);
    ptr_Transport->fn_TX        = fn_TX;
    ptr_Transport->ptr_UserData = ptr_UserData;
}

GlobScop int LPP_RegisterTransport(LPP_Transport_t *ptr_Transport)
{
    if (sys_dnode_is_linked(&ptr_Transport->str_Node))
    {
        return -EALREADY;
    }
    else
    {
        sys_dlist_append(&lpp_str_TransportList, &ptr_Transport->str_Node);
        return 0;
    }
}

GlobScop int LPP_UnregisterTransport(LPP_Transport_t *ptr_Transport)
{
    if (sys_dnode_is_linked(&ptr_Transport->str_Node))
    {
        sys_dlist_remove(&ptr_Transport->str_Node);
        return 0;
    }
    else
    {
        return -EALREADY;
    }
}

GlobScop int LPP_RXFrame(LPP_Transport_t *ptr_Transport, void *ptr_Source, const uint8_t *ptr_Data, size_t siz_Length)
{
    /* Parse the sequence tag: */
    StackScop const char *sz_Sequence = (const char *)ptr_Data;
    StackScop char       *sz_End;
    StackScop uint32_t    u32_Index = (uint32_t)strtoul(sz_Sequence, &sz_End, 10);
    if (*sz_End != '/')
    {
        LOG_ERR("Received invalid sequence tag");
        return -EINVAL;
    }
    sz_Sequence                  = &sz_End[1];
    StackScop uint32_t u32_Count = (uint32_t)strtoul(sz_Sequence, &sz_End, 10);
    if (*sz_End != ':')
    {
        LOG_ERR("Received invalid sequence tag");
        return -EINVAL;
    }

    StackScop size_t siz_SequenceTagLength = (sz_End - (const char *)ptr_Data) + 1;

    LOG_DBG("From %p.%p Got %u/%u %u bytes", ptr_Transport, ptr_Source, u32_Index, u32_Count, siz_Length);

    RTOS_MutexLock(&lpp_str_ActiveMutex, RTOS_FOREVER);
    if (lpp_ptr_ActiveTransport == NULL)
    {
        lpp_ptr_ActiveTransport = ptr_Transport;
        lpp_ptr_ActiveSource    = ptr_Source;
    }

    if (lpp_ptr_ActiveTransport == ptr_Transport && lpp_ptr_ActiveSource == ptr_Source)
    {
        if (u32_Index == lpp_u32_ActiveSequence + 1)
        {
            memcpy(&lpp_arr_RawBuffer[lpp_siz_ActiveLength], &ptr_Data[siz_SequenceTagLength],
                   siz_Length - siz_SequenceTagLength);
            lpp_siz_ActiveLength += siz_Length - siz_SequenceTagLength;
        }
        else
        {
            LOG_ERR("Received out of sequence tag: Got %u, expected %u", u32_Index, lpp_u32_ActiveSequence + 1);
            RTOS_MutexUnlock(&lpp_str_ActiveMutex);
            return -EINVAL;
        }

        if (u32_Index == u32_Count)
        {
            RTOS_SemGive(&lpp_str_ActiveReady);
        }

        RTOS_MutexUnlock(&lpp_str_ActiveMutex);
        return 0;
    }
    else
    {
        RTOS_MutexUnlock(&lpp_str_ActiveMutex);
        return -EBUSY;
    }
}

GlobScop int LPP_Receive(LPP_Request_t *ptr_Request, void *ptr_Destination, size_t *siz_Length,
                         RTOS_Timeout_t tim_Timeout)
{
    size_t siz_BytesAvailable = lpp_siz_ActiveLength - lpp_siz_ActiveReadPosition;
    size_t siz_ToRead         = (*siz_Length < siz_BytesAvailable) ? *siz_Length : siz_BytesAvailable;
    memcpy(ptr_Destination, &lpp_arr_RawBuffer[lpp_siz_ActiveReadPosition], siz_ToRead);
    lpp_siz_ActiveReadPosition += siz_ToRead;
    *siz_Length = siz_ToRead;
    return (lpp_siz_ActiveReadPosition >= lpp_siz_ActiveLength) ? -ENOENT : 0;
}

GlobScop int LPP_Respond(LPP_Request_t *ptr_Request, const ArduinoJson::JsonVariant &ref_Response,
                         size_t siz_ContentLength, RTOS_Timeout_t tim_Timeout)
{
    StackScop StaticJsonDocument<64> cls_Response;
    cls_Response["id"] = ptr_Request->sz_Id;
    if (!ref_Response.isNull())
    {
        cls_Response["body"].shallowCopy(ref_Response);
    }

    StackScop size_t siz_JSONLength = ArduinoJson::measureJson(cls_Response);
    (void)lpp_TXInit(&ptr_Request->str_Response, ptr_Request->ptr_Transport, ptr_Request->ptr_Source,
                     siz_JSONLength + siz_ContentLength, tim_Timeout);
    StackScop TXWriter cls_TXWriter(&ptr_Request->str_Response, tim_Timeout);
    serializeJson(cls_Response, cls_TXWriter);
    return 0;
}

GlobScop int LPP_ResponseSend(LPP_Request_t *ptr_Request, const void *ptr_Content, size_t *siz_Length,
                              RTOS_Timeout_t tim_Timeout)
{
    return lpp_TXSend(&ptr_Request->str_Response, ptr_Content, siz_Length, tim_Timeout);
}
