/**********************************************************************************************************************
* @file            IAL.h
*
* @brief           ICC Abstraction Library. This service acts as an abstraction layer for ICC function calls 
                   and callback registration for applications.
*
* @author          Devan Hone
* @date            07/31/2024
* @copyright       2024 Lion Energy LLC.
*                  All rights reserved
*
*********************************************************************************************************************/
/**
* @addtogroup      IAL
*                  Identifier Description for the ICC Abstraction Library (IAL) Module  
* 
* @ref             IAL.h 
*                  
*********************************************************************************************************************/
#include "COD_Compiler_Defines.h"

#ifndef LION_IAL_H__
#define LION_IAL_H__

/*********************************************************************************************************************
* @cond                                              GLOBAL SCOPE                                             @endcond
*********************************************************************************************************************/

#include <lion/icc/ICC.h>
#include <stdlib.h>

#ifdef __cplusplus
extern "C"
{
#endif

//---------------------------------------//
// Section: Defines/Enums/Typedefs       //
//---------------------------------------//

#define IAL_u16_ICC_REQUEST_BUFFER_SIZE     2048
#define IAL_u16_ICC_RESPONSE_BUFFER_SIZE    1024

#define IAL_u8_ICC_CALL_TIMEOUT_S           10
#define IAL_u8_ICC_CONFIG_CALL_TIMEOUT_S    30
#define IAL_u8_ICC_EVENT_TIMEOUT_S          1

#define IAL_GET_QUANTIFIABLE_DATA_ID    0x02000100
#define IAL_SET_QUANTIFIABLE_DATA_ID    0x02000101
#define IAL_GET_UNIT_DATA_ID            0x02000102
#define IAL_SET_UNIT_DATA_ID            0x02000103
#define IAL_GET_UNIT_CODES_ID           0x02000104
#define IAL_GET_BUFFERED_DATA           0x02000105
#define IAL_READ_REGISTERS_ID           0x02000106
#define IAL_WRITE_REGISTERS_ID          0x02000107
#define IAL_READ_MULTIPLE_ID            0x02000108
#define IAL_WRITE_MULTIPLE_ID           0x02000109
#define IAL_GET_EMS_CONFIG_ID           0x02000200
#define IAL_SET_EMS_CONFIG_ID           0x02000201
#define IAL_GET_PMU_CONFIG_ID           0x03000200
#define IAL_SET_PMU_CONFIG_ID           0x03000201
#define IAL_GET_UNIT_CONFIG_ID          0x02000202
#define IAL_SET_UNIT_CONFIG_ID          0x02000203
#define IAL_GET_TOU_ID                  0x02000204
#define IAL_SET_TOU_ID                  0x02000205
#define IAL_DCU_RESTART_ID              0x4200
#define IAL_PMU_RESTART_ID              0x4202
#define IAL_DCU_REINIT_COM_ID           0x02000300
#define IAL_ALERT_STATUS_ID             0x01000310
#define IAL_VERSION_INFO_ID             0x01000301

#define IAL_sz_UNIT_TYPE_ATS     COD_sz_UNIT_TYPE_ATS
#define IAL_sz_UNIT_TYPE_DCDC    COD_sz_UNIT_TYPE_DCDC
#define IAL_sz_UNIT_TYPE_DCU     COD_sz_UNIT_TYPE_DCU
#define IAL_sz_UNIT_TYPE_EMS     COD_sz_UNIT_TYPE_EMS
#define IAL_sz_UNIT_TYPE_GW      COD_sz_UNIT_TYPE_GW
#define IAL_sz_UNIT_TYPE_HVAC    COD_sz_UNIT_TYPE_HVAC
#define IAL_sz_UNIT_TYPE_PCS     COD_sz_UNIT_TYPE_PCS
#define IAL_sz_UNIT_TYPE_PDU     COD_sz_UNIT_TYPE_PDU
#define IAL_sz_UNIT_TYPE_PMU     COD_sz_UNIT_TYPE_PMU

#define IAL_sz_TYPE_TOU          "tou"

//---------------------------------------//
// Section: Function Prototypes          //
//---------------------------------------//

/**
 * @brief Initialize the ICC Abstraction Library (IAL).
 * 
 * @param ptr_IccCtx A pointer to the ICC context.
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure.
 */
GlobScop int IAL_Init(ICC_Context_t *ptr_IccCtx);

/**
 * @brief  Registers a callback function to be called when the specified function ID is received from the ICC.
 * 
 * @param u32_FunctionID  The function ID the handler will be registered to.
 * @param ptr_str_Handler Pointer to the function handler to be registered.
 * @param fn_Callback     Callback function to be called when an alert status update is received.
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure
 */
GlobScop int IAL_RegisterCallback(uint32_t u32_FunctionID, ICC_Function_Handler_t *ptr_str_Handler,
                                    ICC_Function_HandlerCallback_t fn_Callback);

/**
 * @brief  Unregisters a callback function from the ICC.
 * 
 * @param ptr_str_Handler Pointer to the function handler to be unregistered.
 * 
 * @returns EXIT_SUCCESS or 0 upon success, standard posix errors upon failure.
 */
GlobScop int IAL_UnregisterCallback(ICC_Function_Handler_t *ptr_str_Handler);

#ifdef __cplusplus
}
#endif

#endif // LION_IAL_H__
//EOF
