/**********************************************************************************************************************
* @file            IAL_Modbus.h
*
* @brief           Provides APIs for Modbus direct related ICC functions and events for the CCU.
*                  
*
* @author          Devan Hone
* @date            07/25/2024
* @copyright       2024 Lion Energy LLC.
*                  All rights reserved
*
*********************************************************************************************************************/
/**
* @addtogroup      IAL
*                  Identifier Description for the ICC Abstraction Library (IAL) Module  
* 
* @ref             IAL.h 
*                  
*********************************************************************************************************************/
#include "COD_Compiler_Defines.h"

#ifndef LION_IAL_MODBUS_H__
#define LION_IAL_MODBUS_H__

/*********************************************************************************************************************
* @cond                                              GLOBAL SCOPE                                             @endcond
*********************************************************************************************************************/

#include <lion/ial/IAL.h>

#ifdef __cplusplus
extern "C"
{
#endif

/**
 * @brief   Reads the specified array of register addresses from the modbus device across the ICC.
 * @param   sz_UnitType        "pcs", "dcdc", "hvac", "gw"
 * @param   u8_UnitIndex       (0-2), 255 (Null)
 * @param   arr_u16_Addresses  An array of register addresses to be read.
 * @param   arr_u16_Values     An array of register values to be populated.
 * @param   siz_Request        The number of register addresses to be read.
 * @returns EXIT_SUCCESS or 0 upon success, standard posix error code on failure.
*/
GlobScop int IAL_Mbs_ReadRegisters(const char *sz_UnitType, uint8_t u8_UnitIndex, uint16_t *arr_u16_Addresses,
                                    uint16_t *arr_u16_Values, size_t siz_Request);

/**
 * @brief   Reads a batch of sequential registers from the start address from the modbus device across the ICC for the specified 
 *          size of the request.
 * @param   sz_UnitType        "pcs", "dcdc", "hvac", "gw"
 * @param   u8_UnitIndex       (0-2), 255 (Null)
 * @param   u16_startAddress   The starting register address to be read.
 * @param   arr_u16_Values     An array of register values to be populated.
 * @param   siz_Request        The number of register addresses to be read.
 * @returns EXIT_SUCCESS or 0 upon success, standard posix error code on failure.
*/
GlobScop int IAL_Mbs_ReadMultipleRegisters(const char *sz_UnitType, uint8_t u8_UnitIndex, uint16_t u16_startAddress,
                                            uint16_t *arr_u16_Values, size_t siz_Request);

/**
 * @brief   Writes the array of values to the specified array of register addresses to the modbus device across the ICC.
 * @param   sz_UnitType        "pcs", "dcdc", "hvac", "gw"
 * @param   u8_UnitIndex       (0-2), 255 (Null)
 * @param   arr_u16_Addresses  An array of register addresses to be written.
 * @param   arr_u16_Values     An array of register values to be written.
 * @param   siz_Request        The number of register addresses to be written.
 * @returns EXIT_SUCCESS or 0 upon success, standard posix error code on failure.
*/
GlobScop int IAL_Mbs_WriteRegisters(const char *sz_UnitType, uint8_t u8_UnitIndex, uint16_t *arr_u16_Addresses,
                                    uint16_t *arr_u16_Values, int32_t *arr_s32_results, size_t siz_Request);

/**
 * @brief   Writes a batch of sequential registers from the start address to the modbus device across the ICC for the specified
 *          size of the request.
 * @param   sz_UnitType        "pcs", "dcdc", "hvac", "gw"
 * @param   u8_UnitIndex        (0-2), 255 (Null)
 * @param   u16_startAddress    The starting register address to be written.
 * @param   arr_u16_Values      An array of register values to be written.
 * @param   siz_Request         The number of register addresses to be written.
 * @returns EXIT_SUCCESS or 0 upon success, standard posix error code on failure.

*/
GlobScop int IAL_Mbs_WriteMultipleRegisters(const char *sz_UnitType, uint8_t u8_UnitIndex, uint16_t u16_startAddress,
                                            uint16_t *arr_u16_Values, size_t siz_Request);

#ifdef __cplusplus
}
#endif

#endif /* LION_IAL_MODBUS_H__ */
//EOF
